# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/23 4:10 PM
# @File: __init__.py
# @Email: mlshenkai@163.com
from .transformer import BertClassifier
from .fasttext import FastTextClassifier
from .textcnn import TextCNNClassifier
from .textrnn import TextRNNClassifier
from .ml import ClassicClassifier, TextCluster
