# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/24 2:14 PM
# @File: __init__.py
# @Email: mlshenkai@163.com
from .ml_classifier import ClassicClassifier
from .text_cluster import TextCluster
